#!/usr/bin/env python3

"""
Based on the scores / errors in a learning rate file ("newbob" file),
calculates the best epochs.
"""

from __future__ import annotations

import sys
import argparse

# Returnn imports
import _setup_returnn_env  # noqa

from returnn.util import better_exchook
from returnn.log import log
from returnn.config import Config
from returnn.learning_rate_control import LearningRateControl


def main():
    """
    Main entry point.
    """
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--config", help="RETURNN config")
    arg_parser.add_argument("--learning-rate-file", help="The learning rate file contains scores / errors per epoch.")
    arg_parser.add_argument("--key", help="key to use, e.g. 'dev_error'")
    arg_parser.add_argument("--n", type=int, default=5, help="print best N epochs")
    args = arg_parser.parse_args()

    if bool(args.config) == bool(args.learning_rate_file):
        print("Error: provide either --config or --learning-rate-file")
        arg_parser.print_help()
        sys.exit(1)

    if args.config:
        config = Config()
        config.load_file(args.config)
        lr = LearningRateControl.load_initial_from_config(config)
    elif args.learning_rate_file:
        lr = LearningRateControl(default_learning_rate=1, filename=args.learning_rate_file)  # default lr not relevant
    else:
        assert False, "should not get here with %r" % args

    epochs = sorted(lr.epoch_data.keys())
    if not epochs:
        print("Error: no epochs found")
        sys.exit(1)
    print("Loaded epochs", epochs[0], "..", epochs[-1])

    if args.key:
        key = args.key
        print("Using key %s" % key)
    else:
        last_epoch_with_error_info = None
        for ep in reversed(epochs):
            if lr.epoch_data[ep].error:
                last_epoch_with_error_info = ep
                break
        if last_epoch_with_error_info is None:
            print("Error: no scores/errors found")
            sys.exit(1)
        key = lr.get_error_key(last_epoch_with_error_info)
        print("Using key %s (auto via epoch %i)" % (key, last_epoch_with_error_info))

    epochs_ = []
    missing_epochs = []
    for ep in epochs:
        errors = lr.epoch_data[ep].error
        if key in errors:
            epochs_.append((errors[key], ep))
        else:
            missing_epochs.append(ep)
    if len(epochs_) == len(epochs):
        print("All epochs have the key.")
    else:
        print("Epochs missing the key:", missing_epochs)
    assert epochs_
    epochs_.sort()

    for value, ep in epochs_[: args.n]:
        errors = lr.epoch_data[ep].error
        print(
            ", ".join(
                ["Epoch %i" % ep, "%r %r" % (key, value)] + ["%r %r" % (k, v) for k, v in errors.items() if k != key]
            )
        )


if __name__ == "__main__":
    better_exchook.install()
    log.initialize(verbosity=[5])
    main()



"""
Allow to use Numpy arrays.
This backend will probably never be feature-complete.
It just has the bare minimum such that the user can assign Numpy arrays to Tensor.raw_tensor.
"""

from __future__ import annotations
from typing import Union, Sequence, Tuple
import numpy
from returnn.tensor import Tensor, Dim
from ._backend import Backend


# We do not expect that we will ever implement all the methods of the Backend interface.
# noinspection PyAbstractClass
class NumpyBackend(Backend[numpy.ndarray]):
    """Numpy backend"""

    RawTensorType = numpy.ndarray

    @staticmethod
    def executing_eagerly() -> bool:
        """executing eagerly"""
        return True

    @staticmethod
    def get_dtype_name_raw(raw_tensor: numpy.ndarray) -> str:
        """
        :return: dtype of raw tensor, as string
        """
        return raw_tensor.dtype.name

    @staticmethod
    def as_dtype_raw(dtype_name: str) -> numpy.dtype:
        """
        :param dtype_name: e.g. "float32"
        :return: dtype object
        """
        return numpy.dtype(dtype_name)

    @staticmethod
    def get_ndim_raw(raw_tensor: numpy.ndarray) -> int:
        """
        :return: ndim of raw tensor. assumes it is known
        """
        return raw_tensor.ndim

    @staticmethod
    def get_shape_raw(raw_tensor: numpy.ndarray) -> Tuple[int]:
        """
        :return: shape of raw tensor
        """
        return raw_tensor.shape

    @staticmethod
    def get_shape_tuple_raw(raw_tensor: numpy.ndarray) -> Tuple[int]:
        """
        :return: shape of raw tensor. assumes that ndim is known.
            In eager frameworks, all dims are int.
        """
        return raw_tensor.shape

    @staticmethod
    def get_known_shape_raw(raw_tensor: numpy.ndarray) -> Tuple[int]:
        """
        :return: shape of raw tensor, int for static known, None otherwise. assumes that ndim is known.
            This will not create any ops.
            In eager frameworks, all dims are known.
        """
        return raw_tensor.shape

    @staticmethod
    def reduce(
        source: Tensor[numpy.ndarray],
        *,
        mode: str,
        axis: Union[Dim, Sequence[Dim]],
        use_mask: bool = True,
    ) -> Tensor[numpy.ndarray]:
        """reduce"""
        assert mode in Backend._AllowedReduceModes
        if use_mask:
            if isinstance(axis, Dim):
                assert not axis.need_masking()  # not implemented
            else:
                assert all(not dim.need_masking() for dim in axis)  # not implemented
        func = getattr(numpy, mode)
        raw_dims = (
            [source.get_axis_from_description(axis)]
            if isinstance(axis, Dim)
            else [source.get_axis_from_description(dim) for dim in axis]
        )
        res_dims = [dim for i, dim in enumerate(source.dims) if i not in raw_dims]
        if not res_dims:
            # All are reduced. Need numpy.array() to get a tensor again.
            raw_result = numpy.array(func(source.raw_tensor))
        else:
            raw_result = func(source.raw_tensor, axis=raw_dims)
        res = Tensor(
            name=f"reduce_{mode}",
            raw_tensor=raw_result,
            dims=res_dims,
            dtype=source.dtype,
            sparse_dim=source.sparse_dim,
        )
        return res
